/*-
 * Copyright (c) 2007, Andrea Bittau <a.bittau@cs.ucl.ac.uk>
 *
 */

#ifndef __EASSIDE_COMMON_H__
#define __EASSIDE_COMMON_H__

#define S_DEFAULT_PORT		6969
#define S_DEFAULT_UDP_PORT	6969
#define S_CMD_INET_CHECK        1
#define S_CMD_PACKET		2
#define S_HELLO_LEN		50

#endif /* __EASSIDE_COMMON_H__ */
